# Copyright (c) 2017 NEC Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock

from neutron.common import utils
from oslo_db import exception
from oslo_log import log as logging
from oslo_utils import uuidutils

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.db import models
from networking_odl.journal import dependency_validations
from networking_odl.journal import journal
from networking_odl.tests.unit import base_v2
from networking_odl.tests.unit.db import test_db


class JournalPeriodicProcessorTest(base_v2.OpenDaylightConfigBase):
    @mock.patch.object(journal.OpenDaylightJournalThread, 'set_sync_event')
    def test_processing(self, mock_journal):
        self.cfg.config(sync_timeout=0.1, group='ml2_odl')
        periodic_processor = journal.JournalPeriodicProcessor()
        self.addCleanup(periodic_processor.stop)
        periodic_processor.start()
        utils.wait_until_true(lambda: mock_journal.call_count > 1, 5, 0.1)


class OpenDaylightJournalThreadTest(base_v2.OpenDaylightTestCase):
    def setUp(self):
        super(OpenDaylightJournalThreadTest, self).setUp()
        self.journal = journal.OpenDaylightJournalThread()
        self.addCleanup(self.cleanup)

    @staticmethod
    def cleanup():
        journal.MAKE_URL.clear()

    def test_json_data(self):
        object_type = 'testobject'
        data = 'testdata'
        row = models.OpenDaylightJournal(object_type=object_type,
                                         object_uuid=uuidutils.generate_uuid(),
                                         operation=odl_const.ODL_CREATE,
                                         data=data)

        self.assertEqual("%ss" % object_type, self.journal._json_data(row)[1])

    def test_json_data_customized_url(self):
        object_type = 'randomtestobject'
        data = 'testdata'
        journal.register_url_builder(object_type, lambda row: row.object_type)
        row = models.OpenDaylightJournal(object_type=object_type,
                                         object_uuid=uuidutils.generate_uuid(),
                                         operation=odl_const.ODL_CREATE,
                                         data=data)

        url_param = self.journal._json_data(row)
        self.assertEqual(object_type, url_param[1])

    def test_entry_reset_retries_exceptions(self):
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(
                journal.entry_reset, m, True)

    @mock.patch.object(client.OpenDaylightRestClient, 'sendjson',
                       mock.Mock(side_effect=Exception))
    def test__sync_entry_update_state_by_retry_count_on_exception(self):
        entry = db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        self.journal._max_retry_count = 1
        self.assertEqual(entry.retry_count, 0)
        self.journal._sync_entry(self.db_context, entry)
        self.assertEqual(entry.retry_count, 1)
        self.assertEqual(entry.state, odl_const.PENDING)
        self.journal._sync_entry(self.db_context, entry)
        self.assertEqual(entry.retry_count, 1)
        self.assertEqual(entry.state, odl_const.FAILED)

    def _test__sync_entry_logs(self, log_type):
        entry = db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        logger = self.useFixture(fixtures.FakeLogger())

        self.journal._sync_entry(self.db_context, entry)

        self.assertIn(log_type, logger.output)

    def test__sync_entry_logs_processing(self):
        self._test__sync_entry_logs(journal.LOG_PROCESSING)

    def test__sync_entry_logs_completed(self):
        self._test__sync_entry_logs(journal.LOG_COMPLETED)

    @mock.patch.object(client.OpenDaylightRestClient, 'sendjson',
                       mock.Mock(side_effect=Exception))
    def test__sync_entry_logs_failed(self):
        self._test__sync_entry_logs(journal.LOG_ERROR_PROCESSING)


def _raise_DBReferenceError(*args, **kwargs):
    args = [mock.Mock(unsafe=True)] * 4
    e = exception.DBReferenceError(*args)
    raise e


class JournalTest(base_v2.OpenDaylightTestCase):
    @mock.patch.object(dependency_validations, 'calculate')
    @mock.patch.object(journal.db, 'create_pending_row',
                       side_effect=_raise_DBReferenceError)
    def test_record_triggers_retry_on_reference_error(self, mock_create_row,
                                                      mock_calculate):
        args = [mock.Mock(unsafe=True)] * 5
        self.assertRaises(exception.RetryRequest, journal.record, *args)

    def test_entry_complete_retries_exceptions(self):
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(journal.entry_complete, m, True)

    def _test_entry_complete(self, retention, expected_length):
        self.cfg.config(completed_rows_retention=retention, group='ml2_odl')
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW)
        entry = db.get_all_db_rows(self.db_session)[-1]
        journal.entry_complete(self.db_context, entry)
        rows = db.get_all_db_rows(self.db_session)
        self.assertEqual(expected_length, len(rows))
        self.assertTrue(
            all(row.state == odl_const.COMPLETED for row in rows))

    def test_entry_complete_no_retention(self):
        self._test_entry_complete(0, 0)

    def test_entry_complete_with_retention(self):
        self._test_entry_complete(1, 1)

    def test_entry_complete_with_indefinite_retention(self):
        self._test_entry_complete(-1, 1)

    def test_entry_complete_with_retention_deletes_dependencies(self):
        self.cfg.config(completed_rows_retention=1, group='ml2_odl')
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW)
        entry = db.get_all_db_rows(self.db_session)[-1]
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW,
                              depending_on=[entry])
        dependant = db.get_all_db_rows(self.db_session)[-1]
        journal.entry_complete(self.db_context, entry)
        rows = db.get_all_db_rows(self.db_session)
        self.assertIn(entry, rows)
        self.assertEqual([], entry.dependencies)
        self.assertEqual([], dependant.depending_on)

    def test_entry_reset_retries_exceptions(self):
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(journal.entry_reset, m, True)

    def test_entry_reset(self):
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW)
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW)
        entry = db.get_all_db_rows(self.db_session)[-1]
        entry.state = odl_const.PROCESSING
        self.db_session.merge(entry)
        self.db_session.flush()
        entry = db.get_all_db_rows(self.db_session)[-1]
        self.assertEqual(entry.state, odl_const.PROCESSING)
        journal.entry_reset(self.db_context, entry)
        rows = db.get_all_db_rows(self.db_session)
        self.assertEqual(2, len(rows))
        self.assertTrue(all(row.state == odl_const.PENDING for row in rows))

    def test_entry_set_retry_count_retries_exceptions(self):
        with mock.patch.object(db, 'update_pending_db_row_retry') as m:
            self._test_retry_exceptions(
                journal.entry_update_state_by_retry_count, m, True)

    def test_entry_set_retry_count(self):
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW)
        entry_baseline = db.get_all_db_rows(self.db_session)[-1]
        db.create_pending_row(self.db_session,
                              *test_db.DbTestCase.UPDATE_ROW)
        entry_target = db.get_all_db_rows(self.db_session)[-1]
        self.assertEqual(entry_target.retry_count, 0)
        self.assertEqual(entry_target.retry_count, entry_baseline.retry_count)
        self.assertEqual(entry_target.state, entry_baseline.state)

        journal.entry_update_state_by_retry_count(
            self.db_context, entry_target, 1)
        self.assertEqual(entry_target.retry_count, 1)
        self.assertEqual(entry_target.state, odl_const.PENDING)

        journal.entry_update_state_by_retry_count(
            self.db_context, entry_target, 1)
        self.assertEqual(entry_target.retry_count, 1)
        self.assertEqual(entry_target.state, odl_const.FAILED)
        self.assertNotEqual(entry_target.state, entry_baseline.state)
        self.assertNotEqual(entry_target.retry_count,
                            entry_baseline.retry_count)

    def test_record_logs_recording(self):
        logger = self.useFixture(fixtures.FakeLogger())
        journal.record(self.db_context, *self.UPDATE_ROW)
        for arg in self.UPDATE_ROW[0:3]:
            self.assertIn(arg, logger.output)

    def test_record_logs_dependencies(self):
        entry = db.create_pending_row(self.db_session, *self.UPDATE_ROW)

        logger = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
        journal.record(self.db_context, *self.UPDATE_ROW)
        self.assertIn(str(entry.seqnum), logger.output)



# Copyright (c) 2016 Brocade Communication Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from mock import patch
from neutron.tests import base

from networking_odl.common.client import OpenDaylightRestClient as client
from networking_odl.sfc.flowclassifier import sfc_flowclassifier_v1 as sfc_fc
from networking_odl.tests import base as odl_base
from networking_odl.tests.unit.sfc import constants as sfc_const


class TestOpenDaylightSFCFlowClassifierDriverV1(base.DietTestCase):

    def setUp(self):
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        self.mocked_fc_context = patch(
            'networking_sfc.services.flowclassifier.common.context'
            '.FlowClassifierContext').start().return_value
        super(TestOpenDaylightSFCFlowClassifierDriverV1, self).setUp()

        self.driver = sfc_fc.OpenDaylightSFCFlowClassifierDriverV1()
        self.driver.initialize()
        self.mocked_fc_context.current = sfc_const.FAKE_FLOW_CLASSIFIER

    @patch.object(client, 'sendjson')
    def test_create_flow_classifier(self, mocked_sendjson):
        expected = {"flowclassifier": sfc_const.FAKE_FLOW_CLASSIFIER}
        self.driver.create_flow_classifier(self.mocked_fc_context)
        mocked_sendjson.assert_called_once_with(
            'post', sfc_const.CLASSIFIERS_BASE_URI, expected)

    @patch.object(client, 'sendjson')
    def test_update_flow_classifier(self, mocked_sendjson):
        expected = {"flowclassifier": sfc_const.FAKE_FLOW_CLASSIFIER}
        self.driver.update_flow_classifier(self.mocked_fc_context)
        mocked_sendjson.assert_called_once_with(
            'put', sfc_const.CLASSIFIERS_BASE_URI +
            '/' + sfc_const.FAKE_FLOW_CLASSIFIER_ID, expected)

    @patch.object(client, 'try_delete')
    def test_delete_flow_classifier(self, mocked_try_delete):
        self.driver.delete_flow_classifier(self.mocked_fc_context)
        mocked_try_delete.assert_called_once_with(
            sfc_const.CLASSIFIERS_BASE_URI + '/' +
            sfc_const.FAKE_FLOW_CLASSIFIER_ID)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import threading
import time

from neutron.db import api as db_api
from neutron_lib.callbacks import registry
from neutron_lib import context as nl_context
from neutron_lib.plugins import directory
from neutron_lib import worker
from oslo_config import cfg
from oslo_db import exception
from oslo_log import log as logging
from oslo_service import loopingcall
from requests import exceptions

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.common import utils
from networking_odl.db import db
from networking_odl.journal import dependency_validations


LOG = logging.getLogger(__name__)

MAKE_URL = {}
LOG_ENTRY_TEMPLATE = ("%(log_type)s (Entry ID: %(entry_id)s) - %(op)s "
                      "%(obj_type)s %(obj_id)s")
LOG_RECORDED = 'Recorded'
LOG_PROCESSING = 'Processing'
LOG_COMPLETED = 'Completed'
LOG_ERROR_PROCESSING = 'Error while processing'


def call_thread_on_end(func):
    def new_func(obj, *args, **kwargs):
        return_value = func(obj, *args, **kwargs)
        obj.journal.set_sync_event()
        return return_value
    return new_func


def _enrich_port(plugin_context, ml2_context, object_type, operation, data):
    """Enrich the port with additional information needed by ODL"""

    # NOTE(yamahata): work around of ODL neutron northbound
    # It passes security groups in port as list of dict for historical reasons.
    # keep its format for compatibility.
    # TODO(yamahata): drop this format conversion.
    if data[odl_const.ODL_SGS]:
        groups = [{'id': id_} for id_ in data['security_groups']]
    else:
        groups = []
    new_data = copy.deepcopy(data)
    new_data[odl_const.ODL_SGS] = groups

    # NOTE(yamahata): work around for port creation for router
    # tenant_id=''(empty string) is passed when port is created
    # by l3 plugin internally for router.
    # On the other hand, ODL doesn't accept empty string for tenant_id.
    # In that case, deduce tenant_id from network_id for now.
    # Right fix: modify Neutron so that don't allow empty string
    # for tenant_id even for port for internal use.
    # TODO(yamahata): eliminate this work around when neutron side
    # is fixed
    # assert port['tenant_id'] != ''
    if ('tenant_id' not in new_data or new_data['tenant_id'] == ''):
        if ml2_context:
            network = ml2_context._network_context._network
        else:
            plugin = directory.get_plugin()
            network = plugin.get_network(plugin_context,
                                         new_data['network_id'])
        new_data['tenant_id'] = network['tenant_id']

    return new_data


def _log_entry(log_type, entry, log_level=logging.INFO, **kwargs):
    log_dict = {'log_type': log_type, 'op': entry.operation,
                'obj_type': entry.object_type, 'obj_id': entry.object_uuid,
                'entry_id': entry.seqnum}
    LOG.log(log_level, LOG_ENTRY_TEMPLATE, log_dict, **kwargs)


def record(plugin_context, object_type, object_uuid, operation, data,
           ml2_context=None):
    if (object_type == odl_const.ODL_PORT and
            operation in (odl_const.ODL_CREATE, odl_const.ODL_UPDATE)):
        data = _enrich_port(
            plugin_context, ml2_context, object_type, operation, data)

    # Calculate depending_on on other journal entries
    depending_on = dependency_validations.calculate(
        plugin_context.session, operation, object_type, object_uuid, data)

    # NOTE(mpeterson): Between the moment that a dependency is calculated and
    # the new entry is recorded in the journal, an operation can ocurr that
    # would make the dependency irrelevant. In that case we request a retry.
    # For more details, read the commit message that introduced this comment.
    try:
        entry = db.create_pending_row(
            plugin_context.session, object_type, object_uuid, operation, data,
            depending_on=depending_on)
    except exception.DBReferenceError as e:
        raise exception.RetryRequest(e)

    _log_entry(LOG_RECORDED, entry)
    LOG.debug('Entry with ID %(entry_id)s depends on these entries: '
              '%(depending_on)s',
              {'entry_id': entry.seqnum,
               'depending_on': [d.seqnum for d in depending_on]})


@db_api.retry_if_session_inactive()
def entry_complete(context, entry):
    session = context.session
    with db_api.autonested_transaction(session):
        if cfg.CONF.ml2_odl.completed_rows_retention == 0:
            db.delete_row(session, entry)
        else:
            db.update_db_row_state(session, entry, odl_const.COMPLETED)
            db.delete_dependency(session, entry)


@db_api.retry_if_session_inactive()
def entry_reset(context, entry):
    session = context.session
    with db_api.autonested_transaction(session):
        db.update_db_row_state(session, entry, odl_const.PENDING)


@db_api.retry_if_session_inactive()
def entry_update_state_by_retry_count(context, entry, retry_count):
    session = context.session
    with db_api.autonested_transaction(session):
        db.update_pending_db_row_retry(session, entry, retry_count)


def _make_url(row):
    url_object = utils.make_url_object(row.object_type)
    urlpath = ''
    if row.operation == odl_const.ODL_CREATE:
        urlpath = url_object
    else:
        urlpath = url_object + '/' + row.object_uuid

    return urlpath


def register_url_builder(object_type, method):
    MAKE_URL[object_type] = method


def _build_url(row):
    return MAKE_URL.get(row.object_type, _make_url)(row)


class JournalPeriodicProcessor(worker.BaseWorker):
    """Responsible for running the periodic processing of the journal.

    This is a separate worker as the regular journal thread is called when an
    operation finishes and that run will take care of any and all entries
    that might be present in the journal, including the one relating to that
    operation.

    A periodic run over the journal is thus necessary for cases when journal
    entries in the aforementioned run didn't process correctly due to some
    error (usually a connection problem) and need to be retried.
    """
    def __init__(self):
        super(JournalPeriodicProcessor, self).__init__()
        self._journal = OpenDaylightJournalThread(start_thread=False)
        self._interval = cfg.CONF.ml2_odl.sync_timeout
        self._timer = loopingcall.FixedIntervalLoopingCall(self._call_journal)

    def start(self):
        super(JournalPeriodicProcessor, self).start()
        LOG.debug('JournalPeriodicProcessor starting')
        self._journal.start()
        self._timer.start(self._interval)

    def stop(self):
        LOG.debug('JournalPeriodicProcessor stopping')
        self._timer.stop()

    def wait(self):
        pass

    def reset(self):
        pass

    def _call_journal(self):
        self._journal.set_sync_event()


class OpenDaylightJournalThread(object):
    """Thread worker for the OpenDaylight Journal Database."""

    # make those parameter configurable?
    _RETRY_SLEEP_MIN = 0.1
    _RETRY_SLEEP_MAX = 60

    def __init__(self, start_thread=True):
        self.client = client.OpenDaylightRestClient.create_client()
        self._max_retry_count = cfg.CONF.ml2_odl.retry_count
        self._sleep_time = self._RETRY_SLEEP_MIN
        self.event = threading.Event()
        self._odl_sync_thread = threading.Thread(
            name='sync',
            target=self.run_sync_thread)
        if start_thread:
            self.start()

    def start(self):
        # Start the sync thread
        LOG.debug("Starting a new sync thread")
        self._odl_sync_thread.start()

    def set_sync_event(self):
        self.event.set()

    @staticmethod
    def _json_data(row):
        data = copy.deepcopy(row.data)
        filters.filter_for_odl(row.object_type, row.operation, data)

        if row.operation == odl_const.ODL_CREATE:
            method = 'post'
            to_send = {row.object_type: data}
        elif row.operation == odl_const.ODL_UPDATE:
            method = 'put'
            to_send = {row.object_type: data}
        elif row.operation == odl_const.ODL_DELETE:
            method = 'delete'
            to_send = None

        return method, _build_url(row), to_send

    def run_sync_thread(self):
        while True:
            try:
                self.event.wait()
                self.event.clear()

                self.sync_pending_entries()
            except Exception:
                # Catch exceptions to protect the thread while running
                LOG.exception("Error on run_sync_thread")

    def sync_pending_entries(self):
        LOG.debug("Start processing journal entries")
        context = nl_context.get_admin_context()
        entry = db.get_oldest_pending_db_row_with_lock(context.session)
        if entry is None:
            LOG.debug("No journal entries to process")
            return

        while entry is not None:
            stop_processing = self._sync_entry(context, entry)
            if stop_processing:
                break

            entry = db.get_oldest_pending_db_row_with_lock(context.session)
        LOG.debug("Finished processing journal entries")

    def _retry_sleep(self):
        # When something happened in the connection to ODL, don't busy loop
        # because it's likely to hit same issue.
        # Wait for a while for recovery
        time.sleep(self._sleep_time)
        self._sleep_time = min(self._sleep_time * 2, self._RETRY_SLEEP_MAX)

    def _retry_reset(self):
        self._sleep_time = self._RETRY_SLEEP_MIN

    def _sync_entry(self, context, entry):
        _log_entry(LOG_PROCESSING, entry)
        method, urlpath, to_send = self._json_data(entry)

        # TODO(mkolesni): This logic is weirdly written, need to refactor it.
        try:
            self.client.sendjson(method, urlpath, to_send)
            registry.notify(entry.object_type, odl_const.BEFORE_COMPLETE,
                            self, context=context, operation=entry.operation,
                            row=entry)
            entry_complete(context, entry)
            self._retry_reset()
            _log_entry(LOG_COMPLETED, entry)
        except exceptions.ConnectionError:
            # Don't raise the retry count, just log an error & break
            entry_reset(context, entry)
            LOG.error("Cannot connect to the OpenDaylight Controller,"
                      " will not process additional entries")
            self._retry_sleep()
            return True
        except Exception:
            _log_entry(LOG_ERROR_PROCESSING, entry,
                       log_level=logging.ERROR, exc_info=True)
            entry_update_state_by_retry_count(
                context, entry, self._max_retry_count)

        return False



# Copyright (c) 2016 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from debtcollector import removals

from oslo_config import cfg
from oslo_log import log as logging

from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway import service_drivers
from networking_odl.common import client as odl_client

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')

LOG = logging.getLogger(__name__)

L2GATEWAYS = 'l2-gateways'
L2GATEWAY_CONNECTIONS = 'l2gateway-connections'


@removals.removed_class(
    'OpenDaylightL2gwDriver', version='Queens', removal_version='Rocky',
    message="Usage of V1 drivers is deprecated. Please use V2 instead.")
class OpenDaylightL2gwDriver(service_drivers.L2gwDriver):
    """OpenDaylight L2Gateway Service Driver

    This code is the openstack driver for exciting the OpenDaylight L2GW
    facility.
    """

    def __init__(self, service_plugin, validator=None):
        super(OpenDaylightL2gwDriver, self).__init__(service_plugin, validator)
        self.service_plugin = service_plugin
        self.client = odl_client.OpenDaylightRestClient.create_client()
        LOG.warning(
            "ODL: OpenDaylight L2gateway driver has been deprecated "
            "and will be removed. Switch to driver_v2.")

    @property
    def service_type(self):
        return constants.L2GW

    def create_l2_gateway_postcommit(self, context, l2_gateway):
        LOG.info("ODL: Create L2Gateway %(l2gateway)s",
                 {'l2gateway': l2_gateway})
        request = {'l2_gateway': l2_gateway}
        self.client.sendjson('post', L2GATEWAYS, request)

    def delete_l2_gateway_postcommit(self, context, l2_gateway_id):
        LOG.info("ODL: Delete L2Gateway %(l2gatewayid)s",
                 {'l2gatewayid': l2_gateway_id})
        url = L2GATEWAYS + '/' + l2_gateway_id
        self.client.try_delete(url)

    def update_l2_gateway_postcommit(self, context, l2_gateway):
        LOG.info("ODL: Update L2Gateway %(l2gateway)s",
                 {'l2gateway': l2_gateway})
        request = {'l2_gateway': l2_gateway}
        url = L2GATEWAYS + '/' + l2_gateway['id']
        self.client.sendjson('put', url, request)

    def create_l2_gateway_connection_postcommit(self, context,
                                                l2_gateway_connection):
        LOG.info("ODL: Create L2Gateway connection %(l2gwconn)s",
                 {'l2gwconn': l2_gateway_connection})
        odl_l2_gateway_connection = copy.deepcopy(l2_gateway_connection)
        odl_l2_gateway_connection['gateway_id'] = (
            l2_gateway_connection['l2_gateway_id'])
        odl_l2_gateway_connection.pop('l2_gateway_id')
        request = {'l2gateway_connection': odl_l2_gateway_connection}
        self.client.sendjson('post', L2GATEWAY_CONNECTIONS, request)

    def delete_l2_gateway_connection_postcommit(self, context,
                                                l2_gateway_connection_id):
        LOG.info("ODL: Delete L2Gateway connection %(l2gwconnid)s",
                 {'l2gwconnid': l2_gateway_connection_id})
        url = L2GATEWAY_CONNECTIONS + '/' + l2_gateway_connection_id
        self.client.try_delete(url)
